/**
 * The MIT License (MIT)
 * Copyright (c) 2016-2018 Intel Corporation
 * Copyright (c) dātma, inc™
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of 
 * this software and associated documentation files (the "Software"), to deal in 
 * the Software without restriction, including without limitation the rights to 
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of 
 * the Software, and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all 
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

syntax = "proto2";

option java_package = "org.genomicsdb.model";
option java_outer_classname = "Coordinates";

option go_package = "protobuf/";

message ContigPosition {
  required string contig = 1;
  required int64 position = 2;
}

message GenomicsDBColumn {
  oneof column {
    int64 tiledb_column = 1;
    ContigPosition contig_position = 2;
  }
}

message TileDBColumnInterval {
  required int64 begin = 1;
  required int64 end = 2;
}

message ContigInterval {
  required string contig = 1;
  optional int64 begin = 2;
  optional int64 end = 3;
}

message GenomicsDBColumnInterval {
  oneof interval {
    TileDBColumnInterval tiledb_column_interval = 1;
    ContigInterval contig_interval = 2;
  }
}

message GenomicsDBColumnOrInterval {
  oneof column_or_interval {
    GenomicsDBColumn column = 1;
    GenomicsDBColumnInterval column_interval = 2;
  }
}
