#ifndef TEST_COMMON_H
#define TEST_COMMON_H

extern std::string g_query_json_file;
extern std::string g_loader_json_file;
extern std::string g_golden_output_file;
extern bool g_skip_GT_matching;

#endif
