/* DO NOT EDIT THIS FILE - it is machine generated */
#include <jni.h>
/* Header for class org_genomicsdb_GenomicsDBLibLoader */

#ifndef _Included_org_genomicsdb_GenomicsDBLibLoader
#define _Included_org_genomicsdb_GenomicsDBLibLoader
#ifdef __cplusplus
extern "C" {
#endif
/*
 * Class:     org_genomicsdb_GenomicsDBLibLoader
 * Method:    jniGenomicsDBOneTimeInitialize
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_org_genomicsdb_GenomicsDBLibLoader_jniGenomicsDBOneTimeInitialize
  (JNIEnv *, jclass);

#ifdef __cplusplus
}
#endif
#endif
