/* DO NOT EDIT THIS FILE - it is machine generated */
#include <jni.h>
/* Header for class org_genomicsdb_importer_GenomicsDBImporterJni */

#ifndef _Included_org_genomicsdb_importer_GenomicsDBImporterJni
#define _Included_org_genomicsdb_importer_GenomicsDBImporterJni
#ifdef __cplusplus
extern "C" {
#endif
/*
 * Class:     org_genomicsdb_importer_GenomicsDBImporterJni
 * Method:    jniGetChromosomeIntervalsForColumnPartition
 * Signature: (Ljava/lang/String;I)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_org_genomicsdb_importer_GenomicsDBImporterJni_jniGetChromosomeIntervalsForColumnPartition
  (JNIEnv *, jclass, jstring, jint);

/*
 * Class:     org_genomicsdb_importer_GenomicsDBImporterJni
 * Method:    jniConsolidateTileDBArray
 * Signature: (Ljava/lang/String;Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_org_genomicsdb_importer_GenomicsDBImporterJni_jniConsolidateTileDBArray
  (JNIEnv *, jclass, jstring, jstring);

/*
 * Class:     org_genomicsdb_importer_GenomicsDBImporterJni
 * Method:    jniGenomicsDBImporter
 * Signature: (Ljava/lang/String;I)I
 */
JNIEXPORT jint JNICALL Java_org_genomicsdb_importer_GenomicsDBImporterJni_jniGenomicsDBImporter
  (JNIEnv *, jobject, jstring, jint);

/*
 * Class:     org_genomicsdb_importer_GenomicsDBImporterJni
 * Method:    jniInitializeGenomicsDBImporterObject
 * Signature: (Ljava/lang/String;I)J
 */
JNIEXPORT jlong JNICALL Java_org_genomicsdb_importer_GenomicsDBImporterJni_jniInitializeGenomicsDBImporterObject
  (JNIEnv *, jobject, jstring, jint);

/*
 * Class:     org_genomicsdb_importer_GenomicsDBImporterJni
 * Method:    jniCopyVidMap
 * Signature: (J[B)J
 */
JNIEXPORT jlong JNICALL Java_org_genomicsdb_importer_GenomicsDBImporterJni_jniCopyVidMap
  (JNIEnv *, jobject, jlong, jbyteArray);

/*
 * Class:     org_genomicsdb_importer_GenomicsDBImporterJni
 * Method:    jniCopyCallsetMap
 * Signature: (J[B)J
 */
JNIEXPORT jlong JNICALL Java_org_genomicsdb_importer_GenomicsDBImporterJni_jniCopyCallsetMap
  (JNIEnv *, jobject, jlong, jbyteArray);

/*
 * Class:     org_genomicsdb_importer_GenomicsDBImporterJni
 * Method:    jniAddBufferStream
 * Signature: (JLjava/lang/String;ZJ[BJ)V
 */
JNIEXPORT void JNICALL Java_org_genomicsdb_importer_GenomicsDBImporterJni_jniAddBufferStream
  (JNIEnv *, jobject, jlong, jstring, jboolean, jlong, jbyteArray, jlong);

/*
 * Class:     org_genomicsdb_importer_GenomicsDBImporterJni
 * Method:    jniSetupGenomicsDBLoader
 * Signature: (JLjava/lang/String;)J
 */
JNIEXPORT jlong JNICALL Java_org_genomicsdb_importer_GenomicsDBImporterJni_jniSetupGenomicsDBLoader
  (JNIEnv *, jobject, jlong, jstring);

/*
 * Class:     org_genomicsdb_importer_GenomicsDBImporterJni
 * Method:    jniWriteDataToBufferStream
 * Signature: (JII[BJ)V
 */
JNIEXPORT void JNICALL Java_org_genomicsdb_importer_GenomicsDBImporterJni_jniWriteDataToBufferStream
  (JNIEnv *, jobject, jlong, jint, jint, jbyteArray, jlong);

/*
 * Class:     org_genomicsdb_importer_GenomicsDBImporterJni
 * Method:    jniImportBatch
 * Signature: (J[J)Z
 */
JNIEXPORT jboolean JNICALL Java_org_genomicsdb_importer_GenomicsDBImporterJni_jniImportBatch
  (JNIEnv *, jobject, jlong, jlongArray);

#ifdef __cplusplus
}
#endif
#endif
